// Python - C extension for quantity module.
//
// To compile (Python2.3):
//  gcc -c util_ext.c -I/usr/include/python2.3 -o util_ext.o -Wall -O 
//  gcc -shared util_ext.o  -o util_ext.so	
//
// See the module quantity.py
//
//
// Ole Nielsen, GA 2004
	
#include "Python.h"
#include "Numeric/arrayobject.h"
#include "math.h"

//Shared code snippets
#include "util_ext.h"
#include "quantity_ext.h"



int _compute_gradients(int N,
			double* centroids, 
			double* centroid_values,
			int* number_of_boundaries,
			int* surrogate_neighbours,
			double* a,
			double* b) {
			
  int i, k, k0, k1, k2, index3;
  double x0, x1, x2, y0, y1, y2, q0, q1, q2, det;
  
  
  for (k=0; k<N; k++) {
    index3 = 3*k;
    
    if (number_of_boundaries[k] < 2) {
      //Two or three true neighbours

      //Get indices of neighbours (or self when used as surrogate)	    
      //k0, k1, k2 = surrogate_neighbours[k,:]
      
      k0 = surrogate_neighbours[index3 + 0];
      k1 = surrogate_neighbours[index3 + 1];
      k2 = surrogate_neighbours[index3 + 2];           
      if (k0 == k1 || k1 == k2) return -1;      

      //Get data
      q0 = centroid_values[k0];
      q1 = centroid_values[k1];
      q2 = centroid_values[k2];                    

      x0 = centroids[k0*2]; y0 = centroids[k0*2+1]; 
      x1 = centroids[k1*2]; y1 = centroids[k1*2+1]; 
      x2 = centroids[k2*2]; y2 = centroids[k2*2+1];             

      //Gradient
      _gradient(x0, y0, x1, y1, x2, y2, q0, q1, q2, &a[k], &b[k]);
      
    } else if (number_of_boundaries[k] == 2) {
      //One true neighbour

      //#Get index of the one neighbour
      i=0; k0 = k;
      while (i<3 && k0==k) {
	k0 = surrogate_neighbours[index3 + i];
	i++;
      }
      if (k0 == k) return -1;
      
      k1 = k; //self

      //Get data
      q0 = centroid_values[k0];
      q1 = centroid_values[k1];
	    
      x0 = centroids[k0*2]; y0 = centroids[k0*2+1]; 
      x1 = centroids[k1*2]; y1 = centroids[k1*2+1]; 

      //Gradient
      det = x0*y1 - x1*y0;
      if (det != 0.0) {
	a[k] = (y1*q0 - y0*q1)/det;
	b[k] = (x0*q1 - x1*q0)/det;
      }
    } 
    //    else:
    //        #No true neighbours - 	
    //        #Fall back to first order scheme
    //        pass
  }
  return 0;
}            




/////////////////////////////////////////////////
// Gateways to Python



PyObject *compute_gradients(PyObject *self, PyObject *args) {
  
  PyObject *quantity, *domain;
  PyArrayObject 
    *centroids,            //Coordinates at centroids
    *centroid_values,      //Values at centroids    
    *number_of_boundaries, //Number of boundaries for each triangle
    *surrogate_neighbours, //True neighbours or - if one missing - self
    *a, *b;                //Return values
   
  int dimensions[1], N, err;
  
  // Convert Python arguments to C  
  if (!PyArg_ParseTuple(args, "O", &quantity)) 
    return NULL;

  domain = PyObject_GetAttrString(quantity, "domain");     
  if (!domain) 
    return NULL;  

  //Get pertinent variables
  centroids = (PyArrayObject*) PyObject_GetAttrString(domain, "centroids"); 
  if (!centroids) return NULL;    
  
  centroid_values = (PyArrayObject*) 
    PyObject_GetAttrString(quantity, "centroid_values");   
  if (!centroid_values) return NULL;        
  
  surrogate_neighbours = (PyArrayObject*) 
    PyObject_GetAttrString(domain, "surrogate_neighbours");
  if (!surrogate_neighbours) return NULL;        
  
  number_of_boundaries = (PyArrayObject*) 
    PyObject_GetAttrString(domain, "number_of_boundaries");           
  if (!number_of_boundaries) return NULL;            
  
  N = centroid_values -> dimensions[0];
    
  //Release
  Py_DECREF(domain);      
          
  //Allocate space for return vectors a and b (don't DECREF) 
  dimensions[0] = N;
  a = (PyArrayObject *) PyArray_FromDims(1, dimensions, PyArray_DOUBLE);
  b = (PyArrayObject *) PyArray_FromDims(1, dimensions, PyArray_DOUBLE);  
        

  
  err = _compute_gradients(N,
			   (double*) centroids -> data, 
			   (double*) centroid_values -> data,
			   (int*) number_of_boundaries -> data,
			   (int*) surrogate_neighbours -> data,
			   (double*) a -> data,
			   (double*) b -> data);     
			   
  if (err != 0) {
    PyErr_SetString(PyExc_RuntimeError, "Gradient could not be computed");
    return NULL;
  }
  		     
  
//Return values
  return Py_BuildValue("OO", PyArray_Return(a), PyArray_Return(b)); 
}


/*
PyObject *extrapolate_second_order(PyObject *self, PyObject *args) {



PyObject *compute_gradients(PyObject *self, PyObject *args) {
  
  PyObject *quantity, *domain;
  PyArrayObject 
    *centroids,            //Coordinates at centroids
    *centroid_values,      //Values at centroids    
    *number_of_boundaries, //Number of boundaries for each triangle
    *surrogate_neighbours, //True neighbours or - if one missing - self
    *a, *b;                //Return values
   
  int dimensions[1], N, err;
  
  // Convert Python arguments to C  
  if (!PyArg_ParseTuple(args, "O", &quantity)) 
    return NULL;

  domain = PyObject_GetAttrString(quantity, "domain");     
  if (!domain) 
    return NULL;  

  //Get pertinent variables
  centroids = (PyArrayObject*) PyObject_GetAttrString(domain, "centroids"); 
  if (!centroids) return NULL;    
  
  centroid_values = (PyArrayObject*) 
    PyObject_GetAttrString(quantity, "centroid_values");   
  if (!centroid_values) return NULL;        
  
  surrogate_neighbours = (PyArrayObject*) 
    PyObject_GetAttrString(domain, "surrogate_neighbours");
  if (!surrogate_neighbours) return NULL;        
  
  number_of_boundaries = (PyArrayObject*) 
    PyObject_GetAttrString(domain, "number_of_boundaries");           
  if (!number_of_boundaries) return NULL;            
  
  N = centroid_values -> dimensions[0];
    
  //Release
  Py_DECREF(domain);      
          
  //Allocate space for return vectors a and b (don't DECREF) 
  dimensions[0] = N;
  a = (PyArrayObject *) PyArray_FromDims(1, dimensions, PyArray_DOUBLE);
  b = (PyArrayObject *) PyArray_FromDims(1, dimensions, PyArray_DOUBLE);  
        

  
  err = _compute_gradients(N,
			   (double*) centroids -> data, 
			   (double*) centroid_values -> data,
			   (int*) number_of_boundaries -> data,
			   (int*) surrogate_neighbours -> data,
			   (double*) a -> data,
			   (double*) b -> data);     
			   
  if (err != 0) {
    PyErr_SetString(PyExc_RuntimeError, "Gradient could not be computed");
    return NULL;
  }


  PyObject *quantity, *domain, *Tmp;
  PyArrayObject 
    *qv, //Conserved quantities at vertices 
    *qc, //Conserved quantities at centroids
    *neighbours;
    
  int k, i, n, N, k3;
  double beta; //Safety factor
  double *qmin, *qmax, qn;
  
  // Convert Python arguments to C  
  if (!PyArg_ParseTuple(args, "O", &quantity)) 
    return NULL;

  domain = PyObject_GetAttrString(quantity, "domain");     
  if (!domain) 
    return NULL;  
    
  neighbours = (PyArrayObject*) PyObject_GetAttrString(domain, "neighbours");

  qc = (PyArrayObject*) PyObject_GetAttrString(quantity, "centroid_values"); 
  qv = (PyArrayObject*) PyObject_GetAttrString(quantity, "vertex_values");   
  N = qc -> dimensions[0];
*/
  /*
def extrapolate_second_order(self):        
    """Extrapolate conserved quantities from centroid to
    vertices for each volume using
    second order scheme.
    """
        
    a, b = self.compute_gradients()

    V = self.domain.get_vertex_coordinates()
    qc = self.centroid_values
    qv = self.vertex_values
    
    #Check each triangle
    for k in range(self.domain.number_of_elements):
        #Centroid coordinates            
        x, y = self.domain.centroids[k]
        
        #vertex coordinates
        x0, y0, x1, y1, x2, y2 = V[k,:]
        
        #Extrapolate
        qv[k,0] = qc[k] + a[k]*(x0-x) + b[k]*(y0-y)
        qv[k,1] = qc[k] + a[k]*(x1-x) + b[k]*(y1-y)
        qv[k,2] = qc[k] + a[k]*(x2-x) + b[k]*(y2-y)            
  */
//}    



PyObject *limit(PyObject *self, PyObject *args) {
  
  PyObject *quantity, *domain, *Tmp;
  PyArrayObject 
    *qv, //Conserved quantities at vertices 
    *qc, //Conserved quantities at centroids
    *neighbours;
    
  int k, i, n, N, k3;
  double beta; //Safety factor
  double *qmin, *qmax, qn;
  
  // Convert Python arguments to C  
  if (!PyArg_ParseTuple(args, "O", &quantity)) 
    return NULL;

  domain = PyObject_GetAttrString(quantity, "domain");     
  if (!domain) 
    return NULL;  
    
  neighbours = (PyArrayObject*) PyObject_GetAttrString(domain, "neighbours");

  //Get safety factor beta
  Tmp = PyObject_GetAttrString(domain, "beta"); 
  if (!Tmp) 
    return NULL;
      
  beta = PyFloat_AsDouble(Tmp);  
  
  Py_DECREF(Tmp);    
  Py_DECREF(domain);      

  qc = (PyArrayObject*) PyObject_GetAttrString(quantity, "centroid_values"); 
  qv = (PyArrayObject*) PyObject_GetAttrString(quantity, "vertex_values");   
  N = qc -> dimensions[0];
    
  //Find min and max of this and neighbour's centroid values
  qmin = malloc(N * sizeof(double));
  qmax = malloc(N * sizeof(double));  
  for (k=0; k<N; k++) {
    k3=k*3;
    
    qmin[k] = ((double*) qc -> data)[k];
    qmax[k] = qmin[k];
    
    for (i=0; i<3; i++) {
      n = ((int*) neighbours -> data)[k3+i];
      if (n >= 0) {
	qn = ((double*) qc -> data)[n]; //Neighbour's centroid value
	  
	qmin[k] = min(qmin[k], qn);
	qmax[k] = max(qmax[k], qn);
      }
    }
  }
  
  // Call underlying routine
  _limit(N, beta, (double*) qc -> data, (double*) qv -> data, qmin, qmax);
	 
  free(qmin);
  free(qmax);  
  return Py_BuildValue("");  
}



// Method table for python module
static struct PyMethodDef MethodTable[] = {
  {"limit", limit, METH_VARARGS, "Print out"},    
  {"compute_gradients", compute_gradients, METH_VARARGS, "Print out"},      
  {NULL, NULL, 0, NULL}   /* sentinel */
};

// Module initialisation   
void initquantity_ext(void){
  Py_InitModule("quantity_ext", MethodTable);
  
  import_array();     //Necessary for handling of NumPY structures  
}

